// AT89C52 code for AT89C51/C52/C1051/C1052 programmer.
// Supports variable program data file size.
// Uses fixed delays to determine end of write cycles.
// Delays are dependent upon system speed.
// GET_DATA function works with bidirectional cards.
// Anpassung an Turbo C 3.0
// Zeitschleifen durch Delay ersetzt
// letzte nderung: 15.12.1999

#include <stdio.h>
#include <string.h>
//#include <graph.h>
#include <dos.h>

#define FALSE	0
#define TRUE	-1

//#define PBASE	0x378			// LPT1 I/O addresses
//#define PBASE	0x278			// LPT2 I/O addresses
//#define PDATA	(PBASE+0)
//#define PSTAT	(PBASE+1)
//#define PCTRL	(PBASE+2)

#define CHIPSIZE	8192		// AT89C52 FLASH size

// Four-bit function codes (corresponding to P3.7*:P3.6:P2.7:P2.6).

#define	WRITE_DATA	0x6
#define	READ_DATA	0x4
#define	SELECT_VPP	0x2
#define	WRITE_LOCK_1	0x7
#define	WRITE_LOCK_2	0xb
#define	WRITE_LOCK_3	0xd
#define	CHIP_ERASE	0x9
#define	READ_SIGNATURE	0x8

typedef unsigned char     BYTE;
typedef unsigned int      WORD;
typedef unsigned int      BOOLEAN;

BOOLEAN load_data( char *, BYTE *, int * );
BOOLEAN save_data( char *, BYTE *, int );
void erase( BYTE *, int );
void program( BYTE *, BYTE *, int, int );
void xread( BYTE *, BYTE *, int );
BOOLEAN verify( BYTE *, BYTE *, int );
BOOLEAN blank( BYTE * );
void signature( BYTE * );
void lock( BYTE *, int, int );
void reset( BYTE * );
void set_adfunc( BYTE *, BYTE, WORD );
void set_data( BYTE );
BYTE get_data( void );
void enable_address( BYTE * );
void disable_address( BYTE * );
void enable_data( BYTE * );
void disable_data( BYTE * );
void enable_Vpp( BYTE * );
void disable_Vpp( BYTE * );
void slow_pulse( BYTE * );
void fast_pulse( BYTE * );

WORD pctrl, pdata;		// LPTx control and data port addresses


main(argc, argv)
int argc;
char *argv[];
{
	FILE *fptr;
	int fsize;
	BYTE pgmdata[CHIPSIZE], control = 0;
	char *pch, fname[20];
	WORD far *p_lpt1 = (WORD far *)0x00400008;
	WORD far *p_lpt2 = (WORD far *)0x0040000a;

	if ((argc > 1) && ((pch = strpbrk( argv[1], "12" )) != NULL)) {
		switch (*pch) {
		case '1':			// LPT1
			pdata = *p_lpt1;
			pctrl = *p_lpt1 + 2;
			break;
		case '2':			// LPT2
			pdata = *p_lpt2;
			pctrl = *p_lpt2 + 2;
			break;
		}
		if (pdata == 0) {		// port undefined
			puts( "Specified Parallel Port is not implemented." );
			exit( 255 );
		}
	} else {
		puts( "Parallel Port 1 or 2 must be specified on command line." );
		puts( "Usage: <fname> <LPT1 | LPT2>" );
		exit( 255 );
	}

	while (TRUE) {
		//_clearscreen( _GCLEARSCREEN );
		clrscr();
		// Menu.
		puts( "AT89C52 (noch nicht getestet !)");
		puts( "\t\t    5 Volt Vpp\t    12 Volt Vpp\n" );
		puts( "Chip Erase\t\tC\t\tD\n" );
		puts( "Program from File\tE\t\tF" );
		puts( "Verify against File\tG\t\tG" );
		puts( "Save to File\t\tH\t\tH\n" );
		puts( "Blank Check\t\tI\t\tI\n" );
		puts( "Read Signature\t\tJ\t\tJ\n" );
		puts( "Write Lock Bit 1\tK\t\tL" );
		puts( "Write Lock Bit 2\tM\t\tN" );
		puts( "Write Lock Bit 3\tP\t\tQ\n" );
		puts( "Exit\t\t\t\tX\n\n" );

		printf( "Enter selection: " );
		gets( pch );
		*pch |= 0x20;		// convert first char to lower case

		switch (*pch) {
			case 'c':	// chip erase @5v
				erase( &control, 0 );
				break;
			case 'd':	// chip erase @12v
				erase( &control, 1 );
				break;
			case 'e':	// write chip from file @5v
				printf( "Enter file name: " );
				gets( fname );
				fsize = CHIPSIZE;
				if (load_data( fname, pgmdata, &fsize ))
					program( &control, pgmdata, fsize, 0 );
				else {
					//_clearscreen( _GCLEARSCREEN );
					clrscr();
					puts( "Error opening or reading input data file." );
					puts( "\nPress Enter to continue..." );
					gets( pch );
				}
				break;
			case 'f':	// write chip from file @12v
				printf( "Enter file name: " );
				gets( fname );
				fsize = CHIPSIZE;
				if (load_data( fname, pgmdata, &fsize ))
					program( &control, pgmdata, fsize, 1 );
				else {
					//_clearscreen( _GCLEARSCREEN );
					clrscr();
					puts( "Error opening or reading input data file." );
					puts( "\nPress Enter to continue..." );
					gets( pch );
				}
				break;
			case 'g':	// compare chip contents to file
				printf( "Enter file name: " );
				gets( fname );
				fsize = CHIPSIZE;
				if (load_data( fname, pgmdata, &fsize )) {
					if (!verify( &control, pgmdata, fsize )) {
						puts( "\nPress Enter to continue..." );
						gets( pch );
					}
				} else {
					//_clearscreen( _GCLEARSCREEN );
					clrscr();
					puts( "Error opening or reading input data file." );
					puts( "\nPress Enter to continue..." );
					gets( pch );
				}
				break;
			case 'h':	// save chip contents to file
				printf( "Enter file name: " );
				gets( fname );
				xread( &control, pgmdata, CHIPSIZE );
				if (!save_data( fname, pgmdata, CHIPSIZE )) {
					//_clearscreen( _GCLEARSCREEN );
					clrscr();
					puts( "Error opening or reading output data file." );
					puts( "\nPress Enter to continue..." );
					gets( pch );
				}
				break;
			case 'i':	// verify blank chip
				//_clearscreen( _GCLEARSCREEN );
				clrscr();
				if (blank( &control ))
					puts( "Device is blank" );
				else
					puts( "Device is not blank" );
				puts( "\nPress Enter to continue..." );
				gets( pch );
				break;
			case 'j':	// read signature bytes
				//_clearscreen( _GCLEARSCREEN );
				clrscr();
				signature( &control );
				puts( "\nPress Enter to continue..." );
				gets( pch );
				break;
			case 'k':	// write lock bit 1 @5v
				lock( &control, 1, 0 );
				break;
			case 'l':	// write lock bit 1 @12v
				lock( &control, 1, 1 );
				break;
			case 'm':	// write lock bit 2 @5v
				lock( &control, 2, 0 );
				break;
			case 'n':	// write lock bit 2 @12v
				lock( &control, 2, 1 );
				break;
			case 'p':	// write lock bit 3 @5v
				lock( &control, 3, 0 );
				break;
			case 'q':	// write lock bit 3 @12v
				lock( &control, 3, 1 );
				break;
			case 'x':	// exit program
			default:
				//_clearscreen( _GCLEARSCREEN );
				clrscr();
				exit( 0 );
		}
	}
}


//	Read data from indicated file into specified array.
//	If file is smaller than specified byte count, read entire file
//	and change byte count to reflect smaller file size. Encountering
//	end-of-file before the count is satisfied is not an error.
//	If file is larger than specified byte count, read only the
//	specified number of bytes.

BOOLEAN load_data( fname, store, sptr )
char fname[];
BYTE store[];
int *sptr;
{
	FILE *fptr;
	int nbytes;

	if ((fptr = fopen( fname, "rb" )) == NULL)
		return( FALSE );		// file open failed
	nbytes = fread( store, 1, *sptr, fptr );
	if (ferror( fptr ))
		return( FALSE );		// file read failed
	if (feof( fptr ))			// check for end-of-file
		*sptr = nbytes;			// reduce byte count
	fclose( fptr );
	return( TRUE );
}


//	Write data from specified array into indicated file.
//	Returns a boolean value indicating success or failure.

BOOLEAN save_data( fname, store, bcount )
char fname[];
BYTE store[];
int bcount;
{
	FILE *fptr;

	if ((fptr = fopen( fname, "wb" )) == NULL)
		return( FALSE );		// file open failed
	if (fwrite( store, 1, bcount, fptr ) != bcount)
		return( FALSE );		// file write failed
	fclose( fptr );
	return( TRUE );
}


//	Clear the chip memory to all ones. Suggested before programming.
//	The erase voltage parameter corresponds to 12 volts if one,
//	five volts if zero.

void erase( cptr, vsel )
BYTE *cptr;
int vsel;
{
	int i;

	reset( cptr );				// float signals, Vpp=5v
	set_adfunc( cptr, CHIP_ERASE, 0xfff );	// select function
	enable_address( cptr );			// enable func, PSEN*, PROG*
	//for (i=0; i<10; i++)			// delay
	//	;
	delay(1);

	if (vsel) {
		enable_Vpp( cptr );		// Vpp=12v
		//for (i=0; i<25000; i++)		// delay 15 mS Vpp rise->PROG*
		//	;
		delay(15);
	}

	slow_pulse( cptr );			// apply PROG* pulse
	//for (i=0; i<10; i++)			// delay PROG*->adr/data
	//	;
	delay(1);
	reset( cptr );				// float signals, Vpp=5v

	if (vsel)
		//for (i=0; i<25000; i++)		// delay 15 mS for Vpp fall
		//	;
		delay(15);
}


//	Program the chip with the contents of the specified data array.
//	The specified byte count may be less than the number of bytes in
//	the array or the chip. Programming always begins at address zero.
//	The programming voltage parameter corresponds to 12 volts if one,
//	five volts if zero.
//	The maximum programming time is allotted to each byte write.
//	No attempt is made to determine if a byte programs correctly or
//	if it is already programmed with the desired value.

void program( cptr, data, count, vsel )
BYTE *cptr, data[];
int count, vsel;
{
	WORD addr;
	int i;

	reset( cptr );				// float signals, Vpp=5v
	delay(2);
	set_adfunc( cptr, WRITE_DATA, 0xfff );	// select function
	delay(2);
	enable_address( cptr );			// enable func, PSEN*, PROG*
	enable_data( cptr );			// enable bus before write
	delay(10);
	if (vsel) {
		//for (i=0; i<10; i++)		// delay function->Vpp
		//	;
		delay(1);
		enable_Vpp( cptr );		// Vpp=12v
		//for (i=0; i<25000; i++)		// delay 15 mS Vpp rise->PROG*
		//	;
		delay(15);
	}
	delay(5);
	for (addr=0; addr<count; addr++) {
		set_adfunc( cptr, WRITE_DATA, addr );
						// apply address
		set_data( data[addr] );		// apply data
		//for (i=0; i<10; i++)		// delay adr/data->PROG*
		//	;
		delay(3);
		fast_pulse( cptr );		// apply PROG* pulse
		//for (i=0; i<2500; i++)		// delay 1.5 mS for write
		//	;
		delay(3);
	}

	reset( cptr );				// float signals, Vpp=5v

	if (vsel)
		//for (i=0; i<25000; i++)		// delay 15 mS for Vpp fall
		//	;
		delay(15);
}


//	Read the contents of the chip into the specified data array.
//	The specified byte count may be less than the number of bytes in
//	the chip. Reading always begins at address zero.

void xread( cptr, data, count )
BYTE *cptr, data[];
int count;
{
	BYTE tmp;
	WORD addr;
	int i;

	reset( cptr );				// float signals, Vpp=5v
	set_adfunc( cptr, READ_DATA, 0xfff );	// select function
	enable_address( cptr );			// enable func, PSEN*, PROG*

	for (addr=0; addr<count; addr++) {
		set_adfunc( cptr, READ_DATA, addr );
						// apply address
		for (i=0; i<10; i++)		// delay address->data
			;
		data[addr] = get_data();
	}

	reset( cptr );				// float signals, Vpp=5v
}


//	Compare the contents of the chip to the specified data array.
//	The specified byte count may be less than the number of bytes in
//	the chip. Verification always begins at address zero. Failures are
//	displayed by address, with actual contents and expected contents.
//	A boolean value is returned indicating verify pass/fail.

BOOLEAN verify( cptr, data, count )
BYTE *cptr, data[];
int count;
{
	BYTE tmp;
	BOOLEAN flag=TRUE;
	WORD addr;
	int i;

	reset( cptr );				// float signals, Vpp=5v
	set_adfunc( cptr, READ_DATA, 0xfff );	// select function
	enable_address( cptr );			// enable func, PSEN*, PROG*

	for (addr=0; addr<count; addr++) {
		set_adfunc( cptr, READ_DATA, addr );
						// apply address
		for (i=0; i<10; i++)		// delay address->data
			;
		if ((tmp = get_data()) != data[addr]) {
			if (flag) {		// first time only
				//_clearscreen( _GCLEARSCREEN );
				clrscr();
			}
			printf( "verify fail at %.4X  is %.2X sb %.2X\n", addr, tmp, data[addr] );
			flag = FALSE;
		}
	}

	reset( cptr );				// float signals, Vpp=5v
	return( flag );
}


//	Determine if the chip is erased. Locations of failures are not
//	determined. A boolean value is returned indicating blank/not blank.

BOOLEAN blank( cptr )
BYTE *cptr;
{
	BYTE tmp;
	BOOLEAN flag = TRUE;			// default is blank
	WORD addr;
	int i;

	reset( cptr );				// float signals, Vpp=5v
	set_adfunc( cptr, READ_DATA, 0xfff );	// select function
	enable_address( cptr );			// enable func, PSEN*, PROG*

	for (addr=0; addr<CHIPSIZE; addr++) {
		set_adfunc( cptr, READ_DATA, addr );
						// apply address
		//for (i=0; i<10; i++)		// delay address->data
		//	;
		delay(1);
		if (get_data() != 0xff)		// compare to erased value
			flag = FALSE;		// not blank
	}

	reset( cptr );				// float signals, Vpp=5v
	return( flag );
}


//	Read signature bytes.
//	The first byte is at address 30h, the second at 31h. When set to
//	1Eh and 52h, respectively, they identify an AT89C52. The third
//	byte is at address 32h and indicates a 12 volt part when set to ffh
//	and a five volt part when set to 05h.

void signature( cptr )
BYTE *cptr;
{
	BYTE tmp1, tmp2, tmp3;
	int i;

	reset( cptr );				// float signals, Vpp=5v
	set_adfunc( cptr, READ_SIGNATURE, 0x030 );
						// select function, address
	enable_address( cptr );			// enable func, PSEN*, PROG*
	//for (i=0; i<10; i++)			// delay address->data
	//	;
	delay(1);
	tmp1 = get_data();			// read first byte

	set_adfunc( cptr, READ_SIGNATURE, 0x031 );
						// select function, address
	//for (i=0; i<10; i++)			// delay address->data
	//	;
	delay(1);
	tmp2 = get_data();			// read second byte

	set_adfunc( cptr, READ_SIGNATURE, 0x032 );
						// select function, address
	//for (i=0; i<10; i++)			// delay address->data
	//	;
	delay(1);
	tmp3 = get_data();			// read second byte

	printf( "signature byte 1:  %.2X\n", tmp1 );
	printf( "signature byte 2:  %.2X\n", tmp2 );
	printf( "signature byte 3:  %.2X\n", tmp3 );
	reset( cptr );				// float signals, Vpp=5v
}


//	Write specified lock bit.
//	The voltage parameter corresponds to 12 volts if one,
//	five volts if zero.

void lock( cptr, lbit, vsel )
BYTE *cptr;
int lbit, vsel;
{
	int i;

	reset( cptr );				// float signals, Vpp=5v

	switch (lbit) {				// select function
		case 1:
			set_adfunc( cptr, WRITE_LOCK_1, 0x030 );
						// apply function, address
			break;
		case 2:
			set_adfunc( cptr, WRITE_LOCK_2, 0x030 );
						// apply function, address
			break;
		case 3:
			set_adfunc( cptr, WRITE_LOCK_3, 0x030 );
						// apply function, address
			break;
	}

	enable_address( cptr );			// enable func, PSEN*, PROG*
	//for (i=0; i<10; i++)			// delay
	//	;
	delay(1);

	if (vsel) {
		enable_Vpp( cptr );		// Vpp=12v
		//for (i=0; i<25000; i++)		// delay 15 mS Vpp rise->PROG*
		//	;
		delay(15);
	}

	fast_pulse( cptr );			// apply PROG* pulse
	//for (i=0; i<10; i++)			// delay PROG*->adr/data
	//	;
	delay(1);
	reset( cptr );				// float signals, Vpp=5v

	if (vsel)
		//for (i=0; i<25000; i++)		// delay 15 mS for Vpp fall
		//	;
		delay(15);
}


//	Return programmer hardware to the passive state.
//	Writes zeros into the programmer control latch, which floats the
//	address and data busses and puts 5 volts on Vpp. The programmer
//	address latches are loaded with zeros, the data latch with ones.
//	NOTE: The programmer will not power up correctly with ones left
//	on the LPTx data bus. Also, latches must be addressed before
//	the strobe is generated to avoid glitches.

void reset( cptr )
BYTE *cptr;
{
	outportb( pdata, 0 );		// set up data

	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x0c );		// select low address latch
	outportb( pctrl, 0x0d );		// latch data
	outportb( pctrl, 0x0c );

	outportb( pctrl, 0x00 );		// select high address latch
	outportb( pctrl, 0x01 );		// latch data
	outportb( pctrl, 0x00 );

	outportb( pdata, 0xff );		// set up data

	outportb( pctrl, 0x04 );		// select data latch
	outportb( pctrl, 0x05 );		// latch data
	outportb( pctrl, 0x04 );

//	outp( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data

	*cptr = 0;			// save control latch value
}


//	Write specified values into the programmer address and function
//	code latches. Both address and function must be specified.
//	The most significant bit of the function code is in the control
//	latch, and is written without disturbing the control signals.

void set_adfunc( cptr, func, addr )
BYTE *cptr;
BYTE func;
WORD addr;
{
	outportb( pdata, addr );		// set up low byte of address

	outportb( pctrl, 0x0c );		// select low address latch
	outportb( pctrl, 0x0d );		// latch data
	outportb( pctrl, 0x0c );

	outportb( pdata, (func << 5) | (addr >> 8) );
					// set up high byte of address and
					// low three bits of function
	outportb( pctrl, 0x00 );		// select high address latch
	outportb( pctrl, 0x01 );		// latch data
	outportb( pctrl, 0x00 );

	outportb( pdata, (*cptr = (*cptr & 0xfe) | (func >> 3)) );
					// set up high bit of function
	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}


//	Write specified value into the programmer data latch.

void set_data( outdata )
BYTE outdata;
{
	outportb( pdata, outdata );		// set up output data

	outportb( pctrl, 0x04 );		// select data latch
	outportb( pctrl, 0x05 );		// latch data
	outportb( pctrl, 0x04 );

//	outp( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}


//	Return data present on programmer data lines.
//	Does not first disable programmer output data latch.
//	Some parallel interface cards allow the output data buffer
//	to be disabled by setting high bit five of the control register.
//	If the card does not support this feature, the interface and
//	programmer data buffers must duke it out.

BYTE get_data( void )
{
	BYTE tmp;
	int i;

	outportb( pdata, 0xff );		// set LPTx port data high
	outportb( pctrl, 0x24 );		// disable LPTx port data driver
	outportb( pctrl, 0x26 );		// enable read data buffer

	//for (i=0; i<10; i++)		// delay
	for (i=0; i<100; i++)
		;
	tmp = inp( pdata );		// get data

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data

	return( tmp );
}


//	Enable outputs of programmer address and function latches.
//	Note that PSEN* and PROG* are also enabled.

void enable_address( cptr )
BYTE *cptr;
{
	outportb( pdata, (*cptr |= 0x10) );	// set up data

	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}


//	Disable outputs of programmer address latches.
//	Note that PSEN* and PROG* are also disabled.

void disable_address( cptr )
BYTE *cptr;
{
	outportb( pdata, (*cptr &= ~0x10) );	// set up data

	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}


//	Enable output of programmer data latch.

void enable_data( cptr )
BYTE *cptr;
{
	outportb( pdata, (*cptr |= 0x20) );	// set up data

	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}


//	Disable output of programmer data latch.

void disable_data( cptr )
BYTE *cptr;
{
	outportb( pdata, (*cptr &= ~0x20) );	// set up data

	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}


//	Enable 12 volts on Vpp.
//	Note that Vpp will not immediately reach the specified value.

void enable_Vpp( cptr )
BYTE *cptr;
{
	outportb( pdata, (*cptr |= 0x80) );	// set up data

	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}


//	Return Vpp to 5 volts.
//	Note that Vpp will not immediately reach the specified value.

void disable_Vpp( cptr )
BYTE *cptr;
{
	outportb( pdata, (*cptr &= ~0x80) );	// set up data

	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}


//	Generate 10 mS low-going pulse on PROG*.
//	Note that the software timing loop is system-dependent.

void slow_pulse( cptr )
BYTE *cptr;
{
	int i;

	outportb( pdata, (*cptr |= 0x04) );	// set up data

	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	//for ( i=0; i<15000; i++ )	// approx 10 mS on 33 MHz 80386
	//	;
	delay(10);			// delay

	outportb( pdata, (*cptr &= ~0x04) );	// set up data
	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}


//	Generate 100 uS low-going pulse on PROG*.
//	Note that the software timing loop is system-dependent.

void fast_pulse( cptr )
BYTE *cptr;
{
	int i;

	outportb( pdata, (*cptr |= 0x04) );	// set up data

	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	//for ( i=0; i<150; i++ )		// approx 100 uS on 33 MHz 80386
	for (i=0; i<1500; i++)
		;
				// delay

	outportb( pdata, (*cptr &= ~0x04) );	// set up data
	outportb( pctrl, 0x08 );		// select control latch
	outportb( pctrl, 0x09 );		// latch data
	outportb( pctrl, 0x08 );

	outportb( pctrl, 0x04 );		// control signals inactive
	outportb( pdata, 0 );		// clear data
}
